print("import: 'luigi'")
import luigi

print("import: 'luigi.contrib'")
import luigi.contrib

print("import: 'luigi.contrib.hdfs'")
import luigi.contrib.hdfs

print("import: 'luigi.tools'")
import luigi.tools

print("import: 'luigi.scheduler'")
import luigi.scheduler

